import json
import logging
import os
import re
import time

from PIL import Image
from tesserocr import RIL, PyTessBaseAPI, get_languages, iterate_level

path_dir = os.path.dirname(os.path.abspath(__file__))

logger = logging.getLogger(__name__)


class TessObject:
    """TessObject. Object which comport words meta information from tesseract"""

    def __init__(self, children: list = []) -> None:
        """__init__. Text image object with position and meta information

        :param children: character of words
        :type children: list
        :rtype: None
        """
        self.text = ""
        self.conf = 0
        self.bbox = (0, 0, 0, 0)  # x0,y0,x1,y1
        self.children = children
        self.new_line = False
        self.block = None
        self.para = None

    def __str__(self) -> str:
        """__str__.

        :rtype: str
        """
        return str((self.text, self.bbox, self.new_line, self.block, self.para))

    def __repr__(self) -> str:
        """__repr__.

        :rtype: str
        """
        return str(self)


class Pics:
    """Pics."""

    languages = json.load(open(os.path.join(path_dir, "lang.json")))
    available_languages = get_languages()[1]

    def __init__(self, image: object, lang: str) -> None:
        """__init__. Picture class to be ocrized

        :param image: image to be ocrized
        :type image: object
        :param lang: language to be ocrized
        :type lang: str
        :rtype: None
        """
        self.image = Image.open(image)
        if self.languages.get(lang, None) in self.available_languages:
            self.lang = self.languages[lang]
        else:
            logger.warning(
                str(lang)
                + " not found, available languages are "
                + ", ".join(self.available_languages)
                + ".\nEnglish will be used."
            )
            self.lang = "eng"

    def extract(self) -> list[object]:
        """extract. extract words from image

        :rtype: list[object]
        """
        words = None
        with PyTessBaseAPI(lang=self.lang) as api:

            words = ocrizer(self.image, api)

        return words

    def make_para(self) -> list[tuple]:
        """make_para. make paragraphs from picture

        :rtype: list[tuple]
        """
        words = self.extract()
        pos_paras = make_paragraphs(words)
        paragraphs = []
        for para in pos_paras:
            paragraphs.append((fill_word_para(words, para), para))
        return paragraphs


def ocrizer(image: object, api: PyTessBaseAPI) -> list[object]:
    """ocrizer. Ocrize text

    :param image: image to be ocrized
    :type image: object
    :param api: api to be used
    :type api: PyTessBaseAPI
    :rtype: list[object]
    """

    t = time.time()
    api.SetImage(image)
    api.Recognize()
    ri = api.GetIterator()
    words = get_words_bbox(ri, api=api)

    logger.debug("OCR  ending at " + str(time.time() - t))
    return words


def get_words_bbox(
    ri: object, level: list[object] = [RIL.TEXTLINE, RIL.WORD, RIL.SYMBOL], api=None
) -> list[TessObject]:
    """get_words_bbox. Get words bouding box

    :param ri:
    :type ri: object
    :param level:
    :type level: list[object]
    :param api:
    :rtype: list[TessObject]
    """
    lines = []
    for r in iterate_level(ri, level[1]):

        bbox = r.BoundingBoxInternal(level[1])
        # left, top, right, bottom x,y,w,h
        if bbox is not None:
            text = r.GetUTF8Text(level[1])

            if len(re.sub("\s*", "", text)) > 0:  # text != "" or text != "\n":
                if api is None:
                    children = []
                else:
                    children = []
                    ri_children = api.GetIterator()
                    for rc in iterate_level(ri_children, level[2]):
                        bbox_c = rc.BoundingBoxInternal(level[2])
                        if bbox_c is not None and boxinbox(bbox_c, bbox):
                            child = TessObject()
                            child.text = rc.GetUTF8Text(level[2])
                            child.bbox = bbox_c
                            children.append(child)
                line = TessObject(children=children)
                line.text = text
                line.bbox = [
                    round(bbox[0]),
                    round(bbox[1]),
                    round(bbox[2]),
                    round(bbox[3]),
                ]

                line.conf = r.Confidence(level[1])
                line.new_line = r.IsAtBeginningOf(level[0])
                line.block = r.IsAtBeginningOf(RIL.BLOCK)
                line.para = r.IsAtBeginningOf(RIL.PARA)
                lines.append(line)

    return lines


def make_paragraphs(words: list[TessObject]) -> list[(int, int, int, int)]:
    """make_paragraphs. make paragraphs from ocr result

    :param words:
    :type words: list[TessObject]
    :rtype: list[(int, int, int, int)]
    """
    pos_para = None
    pos_paras = []
    for word in words:
        bbox = word.bbox
        if not word.para:
            if pos_para is None:
                pos_para = bbox
            else:
                pos_para = (
                    min(pos_para[0], bbox[0]),
                    min(pos_para[1], bbox[1]),
                    max(pos_para[2], bbox[2]),
                    max(pos_para[3], bbox[3]),
                )
        else:
            if pos_para is not None:
                pos_paras += [pos_para]
            pos_para = bbox
    return pos_paras


def fill_word_para(
    words: list[object], pos_para: (int, int, int, int)
) -> (str, list[object]):
    """fill_word_para. fill words into paragraphs

    :param words:
    :type words: list[object]
    :param pos_para:
    :type pos_para: (int, int, int, int)
    :rtype: (str, list[object])
    """
    text = ""
    words_para = []
    for word in words:
        if boxinbox(word.bbox, pos_para):
            words_para += [word]
            if word.new_line and text != "":
                text += "\n"
            elif text != "":
                text += " "
            text += word.text
    return text, words_para


def boxinbox(box1: (int, int, int, int), box2: (int, int, int, int)) -> bool:
    """boxinbox. check if a box is inside another bounding box

    :param box1: box to be inside
    :type box1: (int, int, int, int)
    :param box2:
    :type box2: (int, int, int, int)
    :rtype: bool
    """
    x, y, x1, y1 = box1
    x2, y2, x3, y3 = box2
    if x2 <= x <= x3 and y2 <= y <= y3 and x2 <= x1 <= x3 and y2 <= y1 <= y3:
        return True
    else:
        return False



import codecs
import logging
import os.path

from setuptools import Extension, dist, find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


logging.basicConfig()
logger = logging.getLogger()
try:
    import numpy
except ImportError:
    dist.Distribution().fetch_build_eggs(["numpy>=1.19.4"])
    import numpy


setup(
    setup_requires=[
        "setuptools>=18.0",
        "numpy>=1.19.4",
    ],
    entry_points={"console_scripts": ["piclsate = picslate.__main__:script"]},
    name="picslate",
    version=get_version("picslate/__init__.py"),
    author="Thomas Portier",
    description="Picture ocr & translation",
    packages=find_packages(),
    package_dir={"picslate": "picslate"},
    package_data={"picslate": ["font/*", "tesseract/*", "font/*/*"]},
    install_requires=["numpy>=1.19.4", "tesserocr"],
    extra_requires={
        "translate": ["translate"],
        "argostranslate": ["argostranslate"],
        "corrector": ["enchant"],
    },
)



import logging

from .corrector import correct
from .tesseract.picture import Pics
from .translate import translate

logger = logging.getLogger(__name__)


class NoTranslatorFound(Exception):
    """NoTranslatorFound."""

    pass


class NoCorrectorFound(Exception):
    """NoCorrectorFound."""

    pass


class Paragraphs:
    """Paragraphs."""

    def __init__(
        self,
        text: str,
        pos: (int, int, int, int),
        from_lang: str,
        to_lang: str,
        words: list[object] | None = None,
        corrector: bool | str = False,
        translator: str = "auto",
    ) -> None:
        """__init__. Object which comport paragraphs text and meta data

        :param text: paragraph text
        :type text: str
        :param pos: paragraph position
        :type pos: (int, int, int, int)
        :param from_lang: lang of the paragraph
        :type from_lang: str
        :param to_lang: lang to be translated
        :type to_lang: str
        :param words: words inside paragraph
        :type words: list[object] | None
        :param corrector: corrector to be used
        :type corrector: bool | str | callable
        :param translator: translator to be used
        :type translator: str | callable
        :rtype: None
        """
        if callable(corrector):
            self.text = corrector(text, from_lang)
        elif corrector is False:
            self.text = text
        else:
            corrector = correct(corrector)
            if corrector is None:
                raise NoCorrectorFound
            self.text = corrector(text, from_lang)
        self.pos = pos
        self.words = words
        self.from_lang = from_lang
        self.to_lang = to_lang
        self._text_translated = None
        self.translator = translator

    @property
    def text_translated(self) -> str:
        """text_translated.

        :rtype: str
        """
        if self._text_translated is None:
            if callable(self.translator):
                translator = self.translator
            else:
                translator = translate(self.translator)
                if translator is None:
                    raise NoTranslatorFound
            self._text_translated = translator(self.text, self.from_lang, self.to_lang)

        return self._text_translated


def makeParagraphs(
    image: object,
    lang: str,
    to_lang: str,
    corrector: object = False,
    translator: object = "auto",
) -> list[Paragraphs]:
    """makeParagraphs.

    :param image:
    :type image: object
    :param lang:
    :type lang: str
    :param to_lang:
    :type to_lang: str
    :param corrector:
    :type corrector: object
    :param translator:
    :type translator: object
    :rtype: list[Paragraphs]
    """
    picture = Pics(image, lang)
    paragraphs = picture.make_para()
    paras = []
    for (text, words), pos in paragraphs:
        paras += [
            Paragraphs(
                text,
                pos,
                lang[:2],
                to_lang,
                words=words,
                corrector=corrector,
                translator=translator,
            )
        ]
    return paras



import logging
import re

try:
    import enchant
    from enchant.checker import SpellChecker
    from enchant.tokenize import EmailFilter, Filter, URLFilter, WikiWordFilter

    _enchant = True
except ImportError:
    _enchant = False

    class Filter:
        pass


from functools import partial

logger = logging.getLogger(__name__)

escape_char = re.compile(r"\\x[0123456789abcdef]+")
break_space = re.compile(r"\u00a0")
word_ocr_correction = {r"(^|[\n(])¢": r"\1c", r"(^|[\n(])@": r"\1a", "™": r"”"}


def correct(mod: str) -> callable:
    """correct. corrector to be used

    :param mod: which dictionnary to be used, default = aspell
    :type mod: str
    :rtype: callable
    """
    if _enchant is True:
        return partial(corr, mod=mod)
    else:
        return None


def corr(text: str, lang: str, mod: str = "aspell") -> str:
    """corr. words corrector

    :param text: text te be corrected
    :type text: str
    :param lang: text's language
    :type lang: str
    :param mod: dictionnary to be used, default aspell
    :type mod: str
    :rtype: str
    """
    chkr = enchant_corrector(lang, mod=mod)
    return correction(chkr, text)


def enchant_corrector(lang: str, mod: str = "aspell") -> object:
    """enchant_corrector. enchant spell checker

    :param lang:
    :type lang: str
    :param mod: dictionnary to be used, default aspell
    :type mod: str
    :rtype: object
    """
    b = enchant.Broker()

    b.set_ordering("*", mod)
    dict_ = b.request_dict(lang)

    chkr = SpellChecker(
        dict_,
        filters=[
            EmailFilter,
            URLFilter,
            WikiWordFilter,
            UpperCaseFilter,
            CapitalizeFilter,
        ],
    )

    return chkr


class CapitalizeFilter(Filter):
    """Filter skipping over UpperCase word.
    This filter skips any words matching the following regular expression:

           ^([A-Z]\w+[A-Z]+\w+)

    That is, any words that are WikiWords.
    """

    _pattern = re.compile(r"^([A-Z]\w+)")

    def _skip(self, word):
        """_skip.

        :param word:
        """
        if self._pattern.match(word):
            return True
        return False


class UpperCaseFilter(Filter):
    """Filter skipping over UpperCase word.
    This filter skips any words matching the following regular expression:

           ^([A-Z]\w+[A-Z]+\w+)

    That is, any words that are WikiWords.
    """

    _pattern = re.compile(r"^([A-Z]+)")

    def _skip(self, word):
        """_skip.

        :param word:
        """
        if self._pattern.match(word):
            return True
        return False


def correction(chkr: object, txt: str) -> str:
    """correction. enchant words corrector

    :param chkr:
    :type chkr: object
    :param txt:
    :type txt: str
    :rtype: str
    """
    txt = re.sub(escape_char, "", txt)
    txt = re.sub(break_space, " ", txt)
    for key, value in word_ocr_correction.items():
        txt = re.sub(key, value, txt)
    chkr.set_text(txt)
    for err in chkr:
        if len(err.suggest()) > 0:
            sug = err.suggest()[0]
        else:
            sug = ""
        err.replace(sug)

    txt = chkr.get_text()
    return txt
